# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class Metric(object):
    """

    """

    def fold_measure(self, model, folds, grad_needed=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param grad_needed:
        :type grad_needed:
        :return:
        :rtype:
        """
        results = [
            self.measure(
                model,
                train_set=train_set,
                test_set=test_set,
                grad_needed=grad_needed)
            for train_set, test_set in folds
        ]

        if not grad_needed:
            return np.mean(results)

        value, gradient = zip(*results)

        # TODO check this

        return np.mean(value), np.mean(np.vstack(gradient), axis=0)

    @staticmethod
    def measure(model, train_set, test_set=None, grad_needed=False):
        """

        :param model:
        :type model:
        :param train_set:
        :type train_set:
        :param test_set:
        :type test_set:
        :param grad_needed:
        :type grad_needed:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg

from .metric import Metric


class LOOCV(Metric):
    """

    """

    @staticmethod
    def measure(model, train_set, test_set=None, grad_needed=False):
        """

        :param model:
        :type model:
        :param train_set:
        :type train_set:
        :param test_set:
        :type test_set:
        :param grad_needed:
        :type grad_needed:
        :return:
        :rtype:
        """

        if grad_needed:
            raise NotImplementedError("LOOCV gradient")

        if test_set:
            posterior_gp = model.get_posterior(train_set)

            mean = posterior_gp.mean_function.marginalize_mean(
                test_set['X']
            )
            var = posterior_gp.covariance_function.marginalize_covariance(
                test_set['X'],
                only_diagonal=True
            )
            data_set = test_set
        else:
            covariance = model.covariance_function.marginalize_covariance(
                train_set['X']
            )
            l_matrix = model.safe_chol(covariance)

            inv_cov = scipy.linalg.cho_solve(
                (l_matrix, True),
                np.eye(l_matrix.shape[0])
            )

            diag_inv_cov = np.diagonal(inv_cov)[:, None]

            mean = train_set['Y'] - \
                   np.divide(np.dot(inv_cov, train_set['Y']), diag_inv_cov)

            var = np.divide(1.0, diag_inv_cov)
            data_set = train_set

        loocv = -np.sum(
            -0.5 * np.log(var) -
            np.power((data_set['Y'] - mean), 2) / (2.0 * var) -
            0.5 * np.log(2 * np.pi)
        )

        return loocv



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .validation import Validation

class Cross(Validation):
    """

    """
    def __init__(self, n_folds):
        self.n_folds = n_folds

    def get_folds(self, data_set):
        """

        :param data_set:
        :type data_set:
        :return:
        :rtype:
        """
        folds_n = list(range(self.n_folds))
        n = data_set['X'].shape[0]
        order = list(np.random.permutation(n))

        folds = []
        for fold_i in folds_n:
            rel_fold_len = int(np.ceil(float(n) / self.n_folds))
            cut1 = rel_fold_len * fold_i
            cut2 = rel_fold_len * (fold_i + 1)
            folds.append(
                Validation.split_data_set(
                    data_set, order, cut1, cut2
                )
            )

        return folds



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

class Validation(object):
    """

    """
    def get_folds(self, data_set):
        """

        :param data_set:
        :type data_set:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    @staticmethod
    def split_data_set(data_set, order, cut1, cut2):
        """

        :param data_set:
        :type data_set:
        :param order:
        :type order:
        :param cut1:
        :type cut1:
        :param cut2:
        :type cut2:
        :return:
        :rtype:
        """

        train_slice1 = slice(None, cut1, None)
        test_slice = slice(cut1, cut2, None)
        train_slice2 = slice(cut2, None, None)
        train_selection = order[train_slice1] + order[train_slice2]
        test_selection = order[test_slice]

        train_set = {
            name: element[train_selection, :]
            for name, element in data_set.items()
        }

        test_set = {
            name: element[test_selection, :]
            for name, element in data_set.items()
        }

        return train_set, test_set



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from ..parameters import WithParameters


class MeanFunction(WithParameters):
    """

    """
    def __init__(self, hyperparams):

        super(MeanFunction, self).__init__(hyperparams)

    def __copy__(self):

        copyed_object = self.__class__()

        copyed_object.set_hyperparams(self.get_hyperparams())

        return copyed_object

    def marginalize_mean(self, x,
                         dmu_dx_needed=False,
                         dmu_dtheta_needed=False,
                         trans=False):
        """

        :param x:
        :type x:
        :param dmu_dx_needed:
        :type dmu_dx_needed:
        :param dmu_dtheta_needed:
        :type dmu_dtheta_needed:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        mean = self.mean(x)

        result = (mean, )

        if dmu_dx_needed:
            dmu_dx = self.dmu_dx(x)
            result += (dmu_dx, )

        if dmu_dtheta_needed:
            dmu_dtheta = self.dmu_dtheta(x, trans)
            result += (dmu_dtheta, )

        if len(result) == 1:
            result = result[0]

        return result

    def mean(self, x):
        """

        :param x:
        :return:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dmu_dx(self, x):
        """

        :param x:
        :type x:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dmu_dtheta(self, x, trans=False):
        """

        :param x:
        :type x:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from mpl_toolkits.mplot3d import Axes3D

plt.style.use('ggplot')


class Plot(object):

    @staticmethod
    def kernel(kernel, data=None, file_name=None):
        """
        Plot a kernel implementation

        :param kernel:
        :type kernel:
        :param data:
        :type data:
        :param file_name:
        :type file_name:
        :return:
        :rtype:
        """

        if data is None:
            data = np.linspace(-3.0, 3.0, 201)[:, None]
            index_mode = False
        else:
            index_mode = True

        middle_index = int(data.shape[0] * 0.5)
        middle_data = data[middle_index, :][None, :]

        plt.figure(figsize=(11, 4))
        plt.clf()
        plt.suptitle(kernel.__class__.__name__)
        plt.subplot(121)
        covar = kernel.marginalize_covariance(data, middle_data)
        covar = covar.reshape(-1)
        if index_mode:
            half_label = np.arange(0, data.shape[0])
            middle_label = middle_index
        else:
            half_label = data
            middle_label = middle_data[0, 0]
        plt.plot(half_label, covar, linestyle='-', color='#1f77b4', ms=40)
        plt.xlabel("input, x")
        plt.ylabel("covariance, k(x, {})".format(middle_label))

        plt.subplot(122)
        covar = kernel.marginalize_covariance(data, data)
        covar = np.flip(covar, 0)
        if index_mode:
            extent = [0, data.shape[0], 0, data.shape[0]]
        else:
            min_data = np.min(data)
            max_data = np.max(data)
            extent = [min_data, max_data, min_data, max_data]
        plt.imshow(
            covar,
            cmap='Blues',
            interpolation='none',
            extent=extent
        )
        cb = plt.colorbar()
        cb.set_label("covariance, k(x, x')")
        plt.xlabel("input, x")
        plt.ylabel("input, x'")
        if file_name is None:
            plt.show()
        else:
            plt.savefig(file_name, bbox_inches='tight')
            plt.close()

    @staticmethod
    def gp_1d(gp, data, test_data=None, file_name=None, n_samples=0):
        """
        Plot gp in 1D

        :param gp:
        :type gp:
        :param data:
        :type data:
        :param test_data:
        :type test_data:
        :param file_name:
        :type file_name:
        :param n_samples:
        :type n_samples:
        :return:
        :rtype:
        """
        plt.figure(figsize=(5, 4))

        show_data, min_values, max_values, diff_values = \
            Plot._get_plot_bounds(data, test_data)

        plot_points = np.arange(
            min_values[0],
            max_values[0],
            diff_values[0] / 500.0
        )[:, None]

        mean = gp.mean_function.marginalize_mean(
            plot_points)

        var = gp.covariance_function.marginalize_covariance(
            plot_points, only_diagonal=True)

        sdev = 3.0 * np.sqrt(var)

        plt.clf()
        if 0 < n_samples:
            posterior_samples = gp.sample(plot_points, n_samples=n_samples)
            plt.plot(
                plot_points.flatten().tolist(),
                posterior_samples, color='#1f77b4', linewidth=0.5
            )
        plt.plot(
            plot_points.flatten().tolist(),
            mean.flatten().tolist(), color='#1f77b4', linewidth=2
        )
        ax = plt.gca()
        ax.fill_between(
            plot_points.flatten().tolist(),
            (mean - sdev).flatten().tolist(),
            (mean + sdev).flatten().tolist(),
            color='#aec7e8'
        )
        plt.plot(
            show_data['X'].flatten().tolist(),
            show_data['Y'].flatten().tolist(), color='#444444',
            linestyle='None', marker='o', markersize=4
        )
        if test_data is not None:
            plt.plot(
                test_data['X'].flatten().tolist(),
                test_data['Y'].flatten().tolist(), color='#d62728',
                linestyle='None', marker='^', markersize=4
            )

        if file_name is None:
            plt.show()
        else:
            plt.savefig(file_name, bbox_inches='tight')
            plt.close()

    @staticmethod
    def gp_2d(gp, data, test_data=None, resolution=40, file_name=None):
        """
        Plot gp in 2D

        :param resolution:
        :type resolution:
        :param test_data:
        :type test_data:
        :param file_name:
        :type file_name:
        :param data:
        :type data:
        :param gp:
        :type gp:
        :return:
        :rtype:
        """
        show_data, min_values, max_values, _ = \
            Plot._get_plot_bounds(data, test_data)

        # Init grid
        x_grid = np.linspace(min_values[0], max_values[0], resolution)
        y_grid = np.linspace(min_values[1], max_values[1], resolution)
        x_grid, y_grid = np.meshgrid(x_grid, y_grid)
        plot_points = np.vstack((x_grid.flatten(), y_grid.flatten())).T

        mean = gp.mean_function.marginalize_mean(
            plot_points)

        # Init plot
        fig = plt.figure()
        axis = fig.gca(projection='3d')

        # plot data
        axis.scatter(
            show_data['X'][:, 0].tolist(),
            show_data['X'][:, 1].tolist(),
            show_data['Y'].tolist(), s=20, color='#444444', marker='o'
        )

        # plot test data
        if test_data is not None:
            axis.scatter(
                test_data['X'][:, 0].tolist(),
                test_data['X'][:, 1].tolist(),
                test_data['Y'].tolist(), s=20, color='#d62728', marker='^'
            )

        # plot GP
        z_points = mean
        z_points = z_points.reshape(resolution, resolution)
        axis.plot_surface(
            x_grid, y_grid,
            z_points, rstride=1,
            cstride=1, cmap='Blues', linewidth=0,
            antialiased=False, alpha=0.3
        )

        axis.zaxis.set_major_locator(LinearLocator(10))
        axis.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

        if file_name is None:
            plt.show()
        else:
            plt.savefig(file_name, bbox_inches='tight')
            plt.close()

    @staticmethod
    def _get_plot_bounds(data, test_data=None, tail=0.5):
        focus_data = data
        if test_data is not None:
            focus_data = test_data

        points = focus_data['X']
        min_values = np.min(points, axis=0)
        max_values = np.max(points, axis=0)
        diff_values = max_values - min_values
        min_values -= diff_values * tail
        max_values += diff_values * tail
        diff_values = max_values - min_values

        show_i = np.logical_and(
            np.all(min_values < data['X'], axis=1),
            np.all(data['X'] < max_values, axis=1)
        )
        show_data = {
            'X': data['X'][show_i],
            'Y': data['Y'][show_i]
        }

        return show_data, min_values, max_values, diff_values



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import time

import numpy as np
import scipy.optimize as spo

from .fitting_method import FittingMethod


class MultiStart(FittingMethod):
    """

    """

    def __init__(self, obj_fun, ls_method="Powell",
                 max_fun_call=5000, max_ls_fun_call=2000):

        if ls_method in ["Newton-CG", "dogleg", "trust-ncg"]:
            raise NotImplementedError("Hessian not implemented for {}".format(
                ls_method
            ))
        self.grad_needed = ls_method in [
            "CG", "BFGS", "Newton-CG", "L-BFGS-B",
            "TNC", "SLSQP", "dogleg", "trust-ncg"
        ]
        self.bounded_search = ls_method in [
            "L-BFGS-B", "TNC", "SLSQP"
        ]
        self.max_fun_call = max_fun_call
        self.max_ls_fun_call = max_ls_fun_call
        self.ls_method = ls_method
        self.obj_fun = obj_fun

    def fit(self, model, folds):
        """
        Optimize Hyperparameters

        :param model:
        :type model:
        :param folds:
        :type folds:
        :return:
        :rtype:
        """

        def measurement_wrapper(mod_params):
            """
            measurement wrapper to optimize hyperparameters

            :param mod_params:
            :type mod_params:
            :return:
            :rtype:
            """
            current['fun_call'] += 1
            current['ls_fun_call'] += 1

            assert current['fun_call'] < self.max_fun_call,\
                "Funcall limit reached"
            assert current['ls_fun_call'] < self.max_ls_fun_call,\
                "LS Funcall limit reached"

            # Evaluate border if out of bounds
            lower_b, upper_b = zip(*model.get_param_bounds(trans=True))
            dist_b = \
                np.clip(mod_params - np.array(upper_b), a_min=0.0, a_max=None) \
                + \
                np.clip(mod_params - np.array(lower_b), a_min=None, a_max=0.0)
            mod_params = mod_params - dist_b

            try:
                model.set_param_values(
                    mod_params,
                    optimizable_only=True,
                    trans=True
                )
                result = self.obj_fun(
                    model=model,
                    folds=folds,
                    grad_needed=self.grad_needed
                )
                if self.grad_needed:
                    value, gradient = result
                else:
                    value = result
            except np.linalg.linalg.LinAlgError as ex:
                raise ex

            # Penalize if out of bounds
            value += 1e3 * np.max(np.abs(dist_b))

            if value < current['value']:
                current['value'] = value
                current['params'] = model.get_param_values(
                    optimizable_only=False,
                    trans=False
                )

            if self.grad_needed:
                return value, gradient

            return value

        ls_method = self.ls_method
        if self.ls_method == "Powell":
            def mod_powell(fun, x0, args=(), **kwargs):
                """

                :return:
                :rtype:
                """
                rand_perm = np.random.permutation(len(x0))
                direc = np.eye(len(x0))
                direc = direc[rand_perm]

                spo.fmin_powell(fun, x0, args, disp=kwargs['disp'], direc=direc)
            ls_method = mod_powell

        bounds = None
        if self.bounded_search:
            bounds = model.get_param_bounds(trans=True)

        start = time.time()

        log = {
            'fun_calls': 0,
            'improvements': 0,
            'restarts': 0,
            'time': 0,
            'best': {
                'params': [],
                'value': np.inf,
                'ls_fun_call': 0,
                'restart': 0,
                'fun_call': 0
            }
        }

        if model.get_param_n(optimizable_only=True) < 1:
            return log

        while log['fun_calls'] < self.max_fun_call:
            # run optimization
            current = {
                'params': [],
                'value': np.inf,
                'ls_fun_call': 0,
                'fun_call': log['fun_calls']
            }
            x_0 = model.get_param_values(
                optimizable_only=True,
                trans=True
            )
            try:
                spo.minimize(
                    measurement_wrapper,
                    x_0, method=ls_method,
                    jac=self.grad_needed, bounds=bounds,
                    options={
                        'disp': False
                    }
                )
            except (AssertionError, np.linalg.linalg.LinAlgError):
                pass

            log['fun_calls'] = current['fun_call']
            log['restarts'] += 1
            if current['value'] < log['best']['value']:
                log['improvements'] += 1
                log['best']['value'] = current['value']
                log['best']['params'] = current['params']
                log['best']['fun_call'] = current['fun_call']
                log['best']['ls_fun_call'] = current['ls_fun_call']
                log['best']['restart'] = log['restarts']
            if np.random.uniform() < 0.1:
                model.set_params_to_default()
            model.set_params_at_random(trans=True)

        end = time.time()

        log['time'] = end - start

        assert log['best']['params'], "No params were found"

        model.set_param_values(
            log['best']['params'],
            optimizable_only=False,
            trans=False
        )

        model.save_current_as_optimized()

        return log



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .covariance_function import CovarianceFunction
from ..parameters import OptimizableParameter
from ..transformations import LogTransformation
from ..transformations import NoneTransformation


class DotProductFunction(CovarianceFunction):
    """

    """
    def __init__(self, data, is_ard, hyperparams):
        default_lengthscale = 1.0
        default_shift = 1.0
        if is_ard:
            dims = data['X'].shape[1]
            default_lengthscale = np.ones(dims).tolist()
            default_shift = np.ones(dims).tolist()

        hyperparams.append(OptimizableParameter(
            'lengthscale', LogTransformation,
            default_value=default_lengthscale,
            min_value=np.exp(-5), max_value=np.exp(10)
        ))

        hyperparams.append(OptimizableParameter(
            'shift', NoneTransformation,
            default_value=default_shift,
            min_value=1e-5, max_value=1e5
        ))

        super(DotProductFunction, self).__init__(hyperparams)

    def covariance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """

        dot_product = self.dot_product(
            mat_a, mat_b=mat_b,
            only_diagonal=only_diagonal
        )

        covariance = self.dot_product_function(dot_product)

        return covariance

    def dot_product(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Dot product of A and B.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: dot product between solutions of A and B.
        :rtype:
        """
        mat_a = (mat_a - self.get_param_value('shift')) / \
            self.get_param_value('lengthscale')

        if mat_b is None:
            if only_diagonal:
                return np.sum(np.square(mat_a), axis=1)[:, None]
            return np.dot(mat_a, mat_a.T)

        mat_b = (mat_b - self.get_param_value('shift')) / \
            self.get_param_value('lengthscale')

        return np.dot(mat_a, mat_b.T)

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :param mat_b: List of solutions in lines and dimensions in columns.
        :return: 3D array with the gradient in every dimension of X.
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dot_product_function(self, dot_product):
        """

        :param dot_product:
        :type dot_product:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class Parametrizable(object):
    """

    """

    def is_array(self):
        """

        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def set_params_at_random(self, trans=True):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")


class WithParameters(Parametrizable):
    """

    """
    def __init__(self, hyperparameters):
        """

        :param hyperparameters:
        :type hyperparameters:
        """
        self.hyperparameters = hyperparameters

    def is_array(self):
        """

        :return:
        :rtype:
        """

        raise True

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        assert (len(params) == self.get_param_n(
            optimizable_only=optimizable_only)),\
            "length of params is not correct"

        i = 0
        for hyperparameter in self.hyperparameters:
            number_of_params = \
                hyperparameter.get_param_n(optimizable_only=optimizable_only)
            param_slice = slice(i, i + number_of_params)
            hyperparameter.set_param_values(
                params[param_slice],
                optimizable_only=optimizable_only,
                trans=trans
            )
            i += number_of_params

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        for hyperparameter in self.hyperparameters:
            hyperparameter.set_params_to_default(
                optimizable_only=optimizable_only
            )

    def set_params_at_random(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        for hyperparameter in self.hyperparameters:
            hyperparameter.set_params_at_random(trans)

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """

        for hyperparameter in self.hyperparameters:
            hyperparameter.save_current_as_optimized()

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        params = []
        for hyperparameter in self.hyperparameters:
            params += hyperparameter.get_param_values(
                optimizable_only=optimizable_only,
                trans=trans
            )

        return params

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        if not recursive:
            return self.__class__.__name__

        params = []

        for hyperparameter in self.hyperparameters:
            name = self.__class__.__name__
            params += [
                name + "_" + item for item in hyperparameter.get_param_keys(
                    optimizable_only=optimizable_only
                )
            ]

        return params

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        bounds = []

        for hyperparameter in self.hyperparameters:
            bounds += hyperparameter.get_param_bounds(trans)

        return bounds

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        n_optimizable = 0

        for hyperparameter in self.hyperparameters:
            n_optimizable += hyperparameter.get_param_n(
                optimizable_only=optimizable_only
            )

        return n_optimizable

    def get_hyperparams(self):
        """

        :return:
        :rtype:
        """

        return self.hyperparameters

    def set_hyperparams(self, hyperparams):
        """

        :param hyperparams:
        :type hyperparams:
        :return:
        :rtype:
        """

        assert type(hyperparams) is list,\
            "hyperparams must be a list"
        self.hyperparameters = hyperparams

    def get_hyperparam(self, name):
        """

        :param name:
        :type name:
        :return:
        :rtype:
        """
        for hyperparameter in self.hyperparameters:
            hp_name = hyperparameter.get_param_keys(
                recursive=False,
                optimizable_only=False
            )
            if name == hp_name:
                return hyperparameter

    def get_param_value(self, name):
        """

        :param name:
        :type name:
        :return:
        :rtype:
        """
        hyperparam = self.get_hyperparam(name)

        value = hyperparam.get_param_values(optimizable_only=False)

        if hyperparam.is_array():
            return value

        return value[0]


class Parameter(Parametrizable):
    """

    """
    def __init__(self, name, transformation, default_value):
        """

        :param name:
        :type name:
        :param transformation:
        :type transformation:
        :param default_value:
        :type default_value:
        """

        self.name = name
        self.transformation = transformation
        self.default_value = default_value
        self.array = hasattr(self.default_value, "__len__")
        self.current_value = self.default_value

        self.dims = 1
        if self.array:
            self.dims = len(self.default_value)

    def is_array(self):
        """

        :return:
        :rtype:
        """
        return self.array

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        assert len(params) == self.dims, \
            "length of {} is not correct".format(self.name)

        if trans:
            if self.array:
                params = self.transformation.inv_trans(params).tolist()
            else:
                params = self.transformation.inv_trans(params)

        if self.array is False:
            self.current_value = params[0]
        else:
            self.current_value = params

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        self.current_value = self.default_value

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        assert self.current_value is not None, \
            "{} has not been initialized".format(self.name)

        current_value = self.current_value
        if trans:
            current_value = self.transformation.trans(current_value)
            if self.array:
                current_value = current_value.tolist()

        if self.array:
            return current_value
        return [current_value]

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if not recursive:
            return self.name

        if self.dims == 1:
            return [self.name]

        return [
            "{}_d{}".format(self.name, dim) for dim in range(self.dims)
        ]

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        return self.dims


class OptimizableParameter(Parameter):
    """

    """
    def __init__(self, name, transformation, default_value=1.0,
                 min_value=-np.inf, max_value=np.inf,
                 jitter_sd=0.1):
        """

        :param name:
        :type name:
        :param transformation:
        :type transformation:
        :param default_value:
        :type default_value:
        :param min_value:
        :type min_value:
        :param max_value:
        :type max_value:
        :param jitter_sd:
        :type jitter_sd:
        """

        super(OptimizableParameter, self).__init__(
            name, transformation, default_value
        )

        self.min_value = np.float64(min_value)
        self.max_value = np.float64(max_value)

        assert (np.all(self.min_value <= default_value) and
                np.all(default_value <= self.max_value)),\
            "{} is out of bounds".format(self.name)

        self.optimized_value = None

        self.jitter_sd = jitter_sd
        


    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        super(OptimizableParameter, self).set_param_values(
            params, optimizable_only=optimizable_only, trans=trans
        )

        assert (np.all(self.min_value <= self.current_value) and
                np.all(self.current_value <= self.max_value)), \
            "{} is out of bounds".format(self.name)

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        self.optimized_value = None

        super(OptimizableParameter, self).set_params_to_default(
            optimizable_only=optimizable_only
        )

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        min_value = self.min_value
        max_value = self.max_value
        if trans:
            min_value = self.transformation.trans(min_value)
            max_value = self.transformation.trans(max_value)

        if self.optimized_value is not None:
            optimized_value = self.optimized_value
            if trans:
                optimized_value = self.transformation.trans(optimized_value)
            current_value = None
            while not (current_value is not None and
                       np.all(min_value < current_value) and
                       np.all(current_value < max_value)):
                current_value = optimized_value + \
                    np.array(
                        np.random.normal(
                            loc=0.0,
                            scale=self.jitter_sd,
                            size=self.dims
                        )
                    )
        else:
            current_value = np.array(np.random.uniform(
                min_value, max_value, self.dims
            ))

        self.set_param_values(current_value, trans=trans)

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        self.optimized_value = self.current_value

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        min_value = self.min_value
        max_value = self.max_value

        if trans:
            min_value = self.transformation.trans(min_value)
            max_value = self.transformation.trans(max_value)

        return [(min_value, max_value)] * self.dims

    def set_min_value(self, min_value):
        """

        :param min_value:
        :type min_value:
        :return:
        :rtype:
        """
        self.min_value = min_value

    def set_max_value(self, max_value):
        """

        :param max_value:
        :type max_value:
        :return:
        :rtype:
        """
        self.max_value = max_value

    def grad_trans(self, df):
        """

        :param df:
        :type df:
        :return:
        :rtype:
        """

        return self.transformation.grad_trans(self.current_value, df)


class FixedParameter(Parameter):
    """

    """

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if optimizable_only:
            return

        super(FixedParameter, self).set_param_values(
            params, optimizable_only=optimizable_only, trans=trans
        )

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return

        super(FixedParameter, self).set_params_to_default(
            optimizable_only=optimizable_only
        )

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        pass

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        pass

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if optimizable_only:
            return []

        return super(FixedParameter, self).get_param_values(
            optimizable_only=optimizable_only, trans=trans
        )

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return []

        return super(FixedParameter, self).get_param_keys(
            recursive=recursive,
            optimizable_only=optimizable_only
        )

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        return []

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return 0

        return super(FixedParameter, self).get_param_n(
            optimizable_only=optimizable_only
        )



# -*- coding: utf-8 -*-
#
#    Copyright 2018 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .validation import Validation

class Unfold(Validation):
    """

    """
    def __init__(self, fold_len):
        self.fold_len = fold_len

    def get_folds(self, data_set):
        """

        :param data_set:
        :type data_set:
        :return:
        :rtype:
        """

        n = data_set['X'].shape[0]
        order = list(range(n))
        rel_fold_len = int(np.ceil(float(n) * self.fold_len))
        cut1 = n - rel_fold_len
        cut2 = n
        folds = [
            Validation.split_data_set(
                data_set, order, cut1, cut2
            )
        ]

        return folds